/* default style sheet generated by perltidy */
/* body {background: #FFFFFF; color: #000000}
pre { color: #000000; 
      background: #FFFFFF;
      font-family: courier;
    } 
*/
.c  { color: #228B22;} /* comment */
.cm { color: #000000;} /* comma */
.co { color: #000000;} /* colon */
.h  { color: #CD5555; font-weight:bold;} /* here-doc-target */
.hh { color: #CD5555; font-style:italic;} /* here-doc-text */
.i  { color: #00688B;} /* identifier */
.j  { color: #CD5555; font-weight:bold;} /* label */
.k  { color: #8B008B; font-weight:bold;} /* keyword */
.m  { color: #FF0000; font-weight:bold;} /* subroutine */
.n  { color: #B452CD;} /* numeric */
.p  { color: #000000;} /* paren */
.pd { color: #228B22; font-style:italic;} /* pod-text */
.pu { color: #000000;} /* punctuation */
.q  { color: #CD5555;} /* quote */
.s  { color: #000000;} /* structure */
.sc { color: #000000;} /* semicolon */
.v  { color: #B452CD;} /* v-string */
.w  { color: #000000;} /* bareword */

/* /perltidy */


ul.menu {
  border: 1px solid #000;
  background-color: #eee;
  padding: 5px;
  list-style: none;
  padding-left: 0.5em;
}
li.menu {
  border-right: 1px solid #000;
  list-style: none;
  padding-left: 0.5em;
  padding-right: 0.3em;
  display: inline;
}
li.menu_last {
  list-style: none;
  padding-left: 0.5em;
  padding-right: 0.3em;
  display: inline;
}
.menu_selected {
  font-weight: bold;
}
.menu_open {
  font-style: italic;
}

a {
  text-decoration: none;
}
a:hover {
  text-decoration: underline;
}

pre {
  padding: 4px;
  background-color: #eee;
  border: 1px solid #ccc;
}

* > code {
  padding: 1px;
  background-color: #eee;
  border: 1px solid #ddd;
}

pre > code {
  padding: 0px;
  background-color: transparent;
  border: none;
}


.header {
  text-align: center;
}
.logo1 {
  font-weight: bold;
  font-size:36px;
}

a.header, a.header:visited, a.header:active, a.header:hover { color: #000 }

body {
  padding-left: 1.8em;
  padding-right: 1.8em;
  background-color: #FFF;
}

hr {
  height: 1px;
  background-color: #ccc;
  border: none;
}

.footer_date {
  text-align: center;
  font-style: italic;
  font-size: small;
}

.costlist {
  border-spacing: 4px;
}
.costlist > thead > th {
  border-bottom: 1px solid #666;
}
.costlist > tr > td {
  padding-right: 0.5em;
}

/*
.ref_top {
  border-top: 1px solid #666;
}
.ref_precise .ref_subject {
  padding: 0px;
  margin: 0px;
}
*/


dir.top {
  margin: 0.8em;
  padding: 0px;
}

dir.reply {
  margin-top: 0.1em;
  margin-bottom: 0px;
}

dir.toc {
  list-style-type: none;
  margin: 0.2em 0em 0.2em 0em;
  padding-left: 1em;
}

div.toc_box {
  background-color: #f8f4ff;
  padding: 0.1em 0em 1em 0.4em;
  margin: 2em 0em 2em 0em;
}
